#ifndef bqtQBlibStreamHH
#define bqtQBlibStreamHH

#include <qblib/types.hh>

class PrintStreamBase
{
public:
    virtual ~PrintStreamBase() {}
    
    PrintStreamBase& operator<< (const BAS_INT v);
    PrintStreamBase& operator<< (const BAS_LNG v);
    PrintStreamBase& operator<< (const BAS_SGN v);
    PrintStreamBase& operator<< (const BAS_DBL v);
    PrintStreamBase& operator<< (const BAS_STR& v);
    PrintStreamBase& operator<< (const char* v);
    PrintStreamBase& operator<< (const char v);
    
    virtual void Write(const void* buf, unsigned n) = 0;
    
    // DoCRLF is not the same as either '\r' nor '\n'.
    // Only on terminal, '\r' and '\n' both produce the same as DoCRLF().
    // In files, DoCRLF() produces '\r\n' whereas CHR$ produces individual bytes.
    virtual void DoCRLF() = 0;

    virtual unsigned GetXpos() const = 0;  // 0..n-1
    virtual unsigned GetXsize() const = 0; // n. 0 = infinite (files)

    // SPC() actually produces space characters. It does not only move the cursor.
    void DoSPC(int count);
    void DoTAB(int count);
    void DoNextTAB();
private:
    bool CheckFit(int n_columns); // no const, it may print CRLF
    void WriteWithFitCheck(const void* buf, unsigned n);
    void WriteWithFitCheck(const std::string& s)
        { WriteWithFitCheck(s.data(), s.size()); }
};

class InputStreamBase
{
public:
    virtual ~InputStreamBase() {}
    
    // LINE INPUT (console and file)
    virtual void ReadLine(BAS_STR& target) = 0;
    
    // INPUT$  (console and file)
    virtual BAS_STR Input_N(unsigned n) = 0;
    
    virtual void BeginInput() = 0;
    virtual void InputTo(BAS_INT& target) = 0;
    virtual void InputTo(BAS_LNG& target) = 0;
    virtual void InputTo(BAS_SGN& target) = 0;
    virtual void InputTo(BAS_DBL& target) = 0;
    virtual void InputTo(BAS_STR& target) = 0;
    virtual bool EndInputOkp() = 0; // end input, were all fields input?
    /* In console input:
         At call of EndInputOkp, false is returned
         if any of the InputTos caused error or if
         line was not completely emptied by InputTos
       In file input:
         InputTo may span multiple lines, half-read
         line does not matter. EndInputOkp returns false
         only if an InputTo was unsatisfied.
     */
    
    static inline const char* GetRedoMsg() { return "\n?Redo from start;\n"; }
};

extern PrintStreamBase& GET_STREAM_OUTPUT(int filenumber);
extern PrintStreamBase& GET_PRINTER();
extern PrintStreamBase& GET_STDOUT();

extern InputStreamBase& GET_STDIN();
extern InputStreamBase& GET_STREAM_INPUT(int filenumber);

class UsingFormatter
{
public:
    template<typename T>
    void DoParam(PrintStreamBase& out, const T& value);
    /* Note: DoParam() can execute B_ERROR(13) (Type mismatch) */
};


static inline const BAS_STR B_FuncINPUT_S(int n)
    { return GET_STDIN().Input_N(n); }
static inline const BAS_STR B_FuncINPUT_S(int n, int filenumber)
    { return GET_STREAM_INPUT(filenumber).Input_N(n); }

const BAS_STR B_FuncINKEY_S();

#endif
